#region usings
using System;
using System.ComponentModel.Composition;
using VVVV.PluginInterfaces.V2;

using System.Windows.Forms;
using System.Reflection;

using VVVV.Utils.Win32;
using System.Reactive.Linq;
using System.Reactive;
using System.Collections.Generic;
using VVVV.Core.Logging;
#endregion usings

namespace VVVV.Nodes
{
	public sealed class CursorHelper
	{	
		private CursorHelper() {}

	    static readonly CursorHelper instance = new CursorHelper();
		
		static readonly string[] names;
		public static string[] CursorNames { get { return names; } }

		static readonly Cursor[] cursors;
		public static Cursor[] Cursors { get { return cursors; } }
		
	    static CursorHelper()
	    {
	    	var properties = typeof(Cursors).GetProperties();
	    	names = new string[properties.Length];
	    	cursors = new Cursor[properties.Length];
	    	for (int i=0; i<properties.Length; i++)
	    	{
	    		names[i] = properties[i].Name;
	    		cursors[i] = (Cursor)properties[i].GetValue(null);
	    	}
	    }
	} 
	
	#region PluginInfo
	[PluginInfo(Name = "SetCursorType", Category = "System", Help = "Changes the cursor visual for (specified) windows", Author = "woei", AutoEvaluate=true)]
	#endregion PluginInfo
	public class SetCursorTypeNode : Input.WindowMessageNode, IPluginEvaluate
	{
		internal class CursorState
		{
			internal bool Handled;
			internal Cursor LastCursor;
		}
		
		#region fields & pins
		[Input("Handle", DefaultValue = -1)]
		public ISpread<int> FHandle;
		
		[Input("Cursor Type", EnumName="CursorTypes")]
		public ISpread<EnumEntry> FCursorType;
		
		[ImportAttribute()]
		protected VVVV.Core.Logging.ILogger FLogger;
		
		Dictionary<IntPtr,CursorState> FSubclassHandled = new Dictionary<IntPtr, CursorState>();
		#endregion fields & pins
		
		protected override void SubclassCreated(Subclass subclass)
        {
        	FSubclassHandled.Add(subclass.HWnd, new CursorState());
        }

        protected override void SubclassDestroyed(Subclass subclass)
        {
        	FSubclassHandled.Remove(subclass.HWnd);
        }
		
		public override void OnImportsSatisfied()
		{
			EnumManager.UpdateEnum("CursorTypes", "Default", CursorHelper.CursorNames);
			base.OnImportsSatisfied();
		}
		
		protected override void Initialize(IObservable<EventPattern<WMEventArgs>> windowMessages, IObservable<bool> disabled)
        {
        	 windowMessages
        		.Where(e => e.EventArgs.Message == WM.SETCURSOR)
        		.Do(e => HandleCursorNotifications(e.EventArgs))
        		.Subscribe();
        }
		
		protected virtual void HandleCursorNotifications(WMEventArgs e)
		{
			int clientArea = (e.LParam.ToInt32() << 16) >> 16;
			if (clientArea != 1)
				return;
			
			if (FCursorType.SliceCount < 1 || FHandle.SliceCount < 1)
				return;
			
			int sliceCount = Math.Max(FCursorType.SliceCount, FHandle.SliceCount);
			for (int i=0; i<sliceCount; i++)
			{
				if ((FHandle[i] < 0) || (FHandle.IndexOf(e.HWnd.ToInt32()) >-1))
				{
					if (!FSubclassHandled[e.HWnd].Handled)
					{
						FSubclassHandled[e.HWnd].Handled = true;
						FSubclassHandled[e.HWnd].LastCursor = Cursor.Current;
					}
					
					Cursor.Current = CursorHelper.Cursors[FCursorType[i].Index];
					e.Handled = true;
					return;
				}
			}
			
			//was handled, not in use anymore, reset to former state
			if (FSubclassHandled[e.HWnd].Handled)
			{
				FSubclassHandled[e.HWnd].Handled = false;
				Cursor.Current = FSubclassHandled[e.HWnd].LastCursor;
			}
		}
		
		public void Evaluate(int spreadMax) {}
	}
}